require 'code'
require 'draw'
require 'sound'

-- key class
_key = {}
_key.__index = _key
function _key:new(btn,sfx)
    local o = {}
    setmetatable(o,_key)
    o.btn = btn
    o.sfx = sfx
    o.t = 0
    o.pos = 0
    o.hit = false
    return o
end

--[[
checks whether key registers (hit)
'pos' is the cursor state.
    0: on the first button press, the cursor moves.
    1: if button is held down, after time=cfg.wait0, the cursor moves again.
    2: if still held down, the cursor continually moves every time=cfg.wait.
]]
function _key:check(dt)
    local z = self.pos    
    self.t = self.t+dt
    if (z==0 and kp==self.btn) or (z==1 and lk.isDown(self.btn) and self.t>cfg.keywait0) then
        self.pos = self.pos+1
        self.t = 0
        self.hit = true
        return
    end
    if z == 2 and lk.isDown(self.btn) and self.t>cfg.keywait then
        self.t = self.t-cfg.keywait
        self.hit = true
        return
    end
    if not lk.isDown(self.btn) then self.pos = 0 end
    self.hit = false
end

-- input area class
_inp_area = {}
_inp_area.__index = _inp_area
--[[
what is a view?
ex: if list has 4 items but there's only space to show 3,
view=1 -> show 1-3, view=2 -> show 2-4
]]
function _inp_area:new(x,y,nvx,nvy)
    local o = {}
    setmetatable(o,_inp_area)
    -- if x is a number, you can only scroll in y dimension
    o.onx,o.ony = type(x)=='table',type(y)=='table'
    if not o.onx then x = table.rep(x,#y) end
    if not o.ony then y = table.rep(y,#x) end
    o.x,o.y = x,y -- coordinates
    o.ix,o.iy = 1,1 -- index pos
    o.ivx,o.ivy = 1,1 -- index view pos
    o.nx,o.ny = #x,#y -- # pos
    o.nvx = nvx or o.nx -- # view
    o.nvy = nvy or o.ny
    return o
end

-- updates position and view from input
function _inp_area:run()
    local ix,iy = self.ix,self.iy
    if self.ony then
        if key.u.hit then self.iy = self.iy-1 end
        if key.d.hit then self.iy = self.iy+1 end
    end
    if self.onx then
        if key.r.hit then self.ix = self.ix+1 end
        if key.l.hit then self.ix = self.ix-1 end
    end
    if (ix~=self.ix or iy~=self.iy) then
        sound.play(sfx.cursor)
    end
    if kp == btn.ent then
        self.btn = 'ent'
    end
    self.ix,self.ivx = _inp_area.get_pos(self.ix,self.ivx,self.nx,self.nvx)
    self.iy,self.ivy = _inp_area.get_pos(self.iy,self.ivy,self.ny,self.nvy)
end

-- draws cursor
function _inp_area:draw_cursor(x,y)
    x,y = x or 0, y or 0
    draw.cursor(self.x[self.ix-self.ivx+1]+x,self.y[self.iy-self.ivy+1]+y)
end

-- if pos is outside view, move view until you can see pos
function _inp_area.get_pos(i,iv,n,v)
    i = (i-1)%n+1
    if i > iv+v-1 then iv = iv+1-v
    elseif i < iv then iv = i end
    return i,iv
end

-- choice class
_choice = {}
_choice.__index = _choice
function _choice:new(l)
    local o = {}
    setmetatable(o,_choice)
    local ys = table.range(0,fn.h*(#l-1),fn.h)
    o.ia = _inp_area:new(0,ys)
    o.l = l
    o.done = false
    return o
end

-- display list and cursor
function _choice:run(x,y)
    local i
    self.ia:run()
    self.ia:draw_cursor(x,y)
    for i = 1,self.ia.ny do
        draw.text(self.l[i],x+img.cursor:getWidth()+cfg.cursordx,y+fn.h*(i-1)+cfg.cursordy)
    end
    if self.ia.btn == 'ent' then
        sound.play(sfx.confirm)
        self.done = true
        self.pick = self.ia.iy -- the choice you picked (a number)
    end
end
