require 'code'
require 'input'
require 'text'

-- load variables
function love.load()
    -- shortcuts for love stuff
    la = love.audio
    le = love.event
    lg = love.graphics
    lk = love.keyboard
    ls = love.system
    lt = love.timer
    
    -- font
    fn = {}
    fn.f = lg.newImageFont('fonts/font.png'," abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,!?-+/():;%&`'*#=[]\"")
    fn.h = fn.f:getHeight()+2
    fn.w = function(s) return fn.f:getWidth(s) end
    lg.setFont(fn.f)
    
    -- colors
    col = {
        white = {255,255,255},
        black = {0,0,0},
        gray = {128,128,128},
        name = {255,255,0},
        shadow = {0,0,0,192},
        window = {0,80,160},
        mblue = {128,192,255},
        mred = {255,128,128}
    }
    
    -- config
    cfg = {
        msgspd = 50, -- chars/sec
        txtshdw = true,
        keywait0 = .4, -- sec
        keywait = .1,
        cursordx = 3, -- delta x,y of text wrt cursor right edge
        cursordy = -2,
        tblink = 0.5 -- sec, msg blinker time period
    }
    
    -- keys
    btn = {
        ent = 'z',
        esc = 'x'
    }
    key = {
        u = _key:new('up'),
        r = _key:new('right'),
        d = _key:new('down'),
        l = _key:new('left')
    }
    kp = '' -- last key press
    
    -- images
    img = {
        cursor = lg.newImage('pics/cursor.png'),
        press = lg.newImage('pics/press.png'),
        face = lg.newImage('pics/face.png')
    }
    
    -- sound effects
    sfx = {
        cursor = la.newSource('sounds/cursor.wav','static'),
        confirm = la.newSource('sounds/confirm.wav','static')
    }
    
    -- other
    dt = 0
    fps = 0
    
    -- for game
    t = {}
    local i    
    s = 'Hi everyone!|,|, This is a message system I\'ve developed for the awesome game engine |c{mblue}LOVE|c{white}.'
    m = _msg:new(s,200,{x=160,y=120,alx='m',alxb='m',alyb='m',box=false})
    table.insert(t,m)
    s = 'It was designed with RPGs in mind, but is suited for any game genre.|n|nI will make a |c{mred}readme|c{white} file for how to use the various features.'
    m = _msg:new(s,160,{})
    table.insert(t,m)
    s = 'There are a few more features I want to add, like scrolling text and a message box animation, but for the most part it\'s complete.'
    m = _msg:new(s,160,{name='litearc',face=img.face,wait=6,skip=false})
    table.insert(t,m)
    s = 'You can create long messages that use the same message box without having to create each message separately. This may be useful, for example, in RPGs, where there is a lot of dialogue. I still need to add in a formatting command to start a new message box as well as manually set the box width.'
    m = _msg.getms(s,160,2,{x=120,y=180,alx='r'})
    for i=1,#m do table.insert(t,m[i]) end
    s = 'So, this is pretty cool right?'
    m = _msg:new(s,160,{name='litearc',choice={'Yea!','Er... sure'}})
    table.insert(t,m)
    
    t1 = {}
    s = 'Thanks! =D'
    m = _msg:new(s,nil,{x=160,y=120,alx='m',alxb='m',alyb='m',box=false,blinker=false})
    table.insert(t1,m)
    
    t2 = {}
    s = 'Okay... =('
    m = _msg:new(s,nil,{x=160,y=120,alx='m',alxb='m',alyb='m',box=false,blinker=false})
    table.insert(t2,m)
    
    tb = {}
    s = 'Anyway, that\'s it!|! Just wanted to show you guys the latest progress.'
    m = _msg:new(s,120,{name='litearc'})
    table.insert(tb,m)
    
    --[[
    s = 'Anyway, feel free to mess around with the library.|! And let me know if you find any bugs!'
    m = _msg:new(s,120,{name='litearc'})
    table.insert(tb,m)
    s = 'And |c{mred}Ref|c{white}, I\'ll try to make the implementation cleaner, so don\'t spend too much time messing around with this...|,|,|nI just wanted to show you the current state of the project since you seemed interested in it. Have fun!'
    m = _msg:new(s,240,{x=40,y=100,name='litearc'})
    table.insert(tb,m)
    ]]
end

-- key press
function love.keypressed(k)
    kp = k
    if k == 'q' then
        le.push('quit')
    end
end

-- update game
function love.update(a)
    dt = a
    local k,_
    for k,_ in pairs(key) do
        key[k]:check(a)
    end
    fps = (fps+1/a)/2
end

-- draw stuff
function love.draw()
    if t1[#t1].over or t2[#t2].over then
        _msg.playlist(tb,20,20)
    end
    if t[#t].over then
        if t[#t].pick == 1 then _msg.playlist(t1,20,20)
        elseif t[#t].pick == 2 then _msg.playlist(t2,20,20) end
    end
    _msg.playlist(t,20,20)
    kp = ''
    lg.setColor(col.white)
    lg.print(math.floor(fps),2,2)
end