require 'code'

--[[
a bloc is text of one color and with no line breaks.
messages are played out in blocs to improve performance.
]]
_bloc = {}
_bloc.__index = _bloc
function _bloc:new(s,x,y,c)
    local o = {}
    setmetatable(o,_bloc)
    o.s = string.unformat(s)
    o.nc = o.s:len()
    o.x,o.y = x,y
    o.c = c or col.white
    local _,r,g,b,a,co
    local n,e = 0,1
    -- parse out and set color
    if s:sub(1,2) == '|c' then
        _,_,r,g,b,a = s:find('(%x%x)(%x%x)(%x%x)(%x%x)',3)
        co = {tonumber(r,16),tonumber(g,16),tonumber(b,16),tonumber(a,16)}
        n = n+10
        o.c = co
    end
    return o
end

-- draw first 'nc' characters of bloc
function _bloc:draw(x,y,nc)
    local nc = nc or -1 -- if nc>s:len(), displays all of s, so don't worry about it
    draw.text(self.s:sub(1,nc),x+self.x,y+self.y,self.c)
end

-- message class
_msg = {}
_msg.__index = _msg
function _msg:new(s,wo,opt)
    local o = {}
    setmetatable(o,_msg)
    
    -- a ton of options!
    o.x = opt.x or nil -- preset x disp pos
    o.y = opt.y or nil  -- preset y disp pos
    o.alx = opt.alx or 'l' -- text x-alignment
    o.alxb = opt.alxb or 'l' -- box x-alignment
    o.alyb = opt.alyb or 't' -- box y-alignment
    o.wait = opt.wait or nil -- if nil, keypress ends msg, else wait opt.wait sec
    if opt.skip == false then o.skip = false else o.skip = true end -- if true, can skip msg with keypress
    if opt.box == false then o.box = false else o.box = true end -- display message box?
    o.chs = opt.choice or nil -- a list of of choices (strings) at end of msg
    o.name = opt.name or nil -- show name at top?
    o.face = opt.face or nil -- show facepic?
    if opt.blinker == false then o.blinker = opt.blinker else o.blinker = true end -- show blinker?
    
    local w = wo
    -- these options are autoset if necessary, do not set them!
    o.facedx = opt.facedx or 0 -- x-shift if using facepic
    o.msgdy = opt.msgdy or (((s=='') and -fn.h) or 0) -- if msg is empty ('') and there are choices, move choices up one line
    if o.face then
        o.wface = o.face:getWidth()
        o.hface = o.face:getHeight()
        if w then
            o.facedx = o.wface+10
            w = w-o.facedx
        end
    end

    s = _msg.replcol(s)
    -- add newlines if width is specified
    if w then s = _msg.addnls(s,w) end
    if s:len() == 0 then s = s..' ' end -- hack to not prevent code crash
    
    --[[ adds pauses in a hackish way. it adds spaces before the pause command
    and adjusts the block display pos. n spaces = pause of n/msgspd sec ]]
    local ps = string.rep(' ',10)
    local pl = string.rep(' ',50)
    s = s:gsub('|,',ps..'|,')
    s = s:gsub('|:',pl..'|:')
    o.b = {}
    -- o.us = string.unformat(s)
    -- o.tt = (#o.us)/cfg.msgspd -- total msg time
    
    local i = 1 -- i:pos in 's'
    local j = 1 -- pos where last bloc ended
    local e = 1 -- current bloc #
    local cl = 1 -- current line #
    o.lw = {} -- line widths
    o.bn = {} -- line # of each bloc

    local xc,yc = 0,0 -- used to keep track of block pos
    local dx,dy
    local cc = col.white -- current text color
    local b -- stores each bloc as it is parsed
    local z
    while true do
        b = ''
        z = s:sub(i,i+1)
        
        -- a new bloc starts when a formatting command is encountered
        if z == '|!' then
            b = s:sub(j,i-1)
            dx,dy = fn.w(string.unformat(b)),0
        elseif z == '|c' then
            b = s:sub(j,i-1)
            dx,dy = fn.w(string.unformat(b)),0
        elseif z == '|n' then
            b = s:sub(j,i-1)
            o.lw[cl] = xc+fn.w(string.unformat(b))
            o.bn[cl] = e
            cl = cl+1
            dx,dy = -xc,fn.h
        elseif z == '|,' then
            b = s:sub(j,i-1)
            dx,dy = -fn.w(ps)+fn.w(string.unformat(b)),0
        elseif z == '|:' then
            b = s:sub(j,i-1)
            dx,dy = -fn.w(pl)+fn.w(string.unformat(b)),0
        end
        
        -- reached the end        
        if i >= s:len() then
            b = s:sub(j,i)
            dx,dy = fn.w(string.unformat(b)),0
            o.lw[cl] = xc+dx
            o.bn[cl] = e
        end
        
        -- create bloc if necessary
        if b ~= '' then
            o.b[e] = _bloc:new(b,xc,yc,cc)
            cc = o.b[e].c
            
            -- keypress stuff
            o.b[e].inp = (z == '|!')
            if (z == '|!') then
                -- o.b[e].t0 = 0
                o.b[e].tset = false
                o.b[e].done = false
            end
            
            xc,yc = xc+dx,yc+dy
            e,j = e+1,i
        end
        
        if i >= s:len() then break end
        i = i+1
        
    end
    
    -- alignment
    o.nl = cl -- total number of lines (not including choices)
    if o.chs then
        o.nchs = #o.chs
        o.cchs = _choice:new(o.chs)
        -- set box width to longest line width (may be a choice)
        local chlw = {} -- choice line widths
        for i = 1,o.nchs do chlw[i] = fn.w(o.chs[i])+img.cursor:getWidth()+3 end
        o.wll = math.max(unpack(o.lw),unpack(chlw)) -- width of longest line
    else
        o.nchs = 0
        o.wll = math.max(unpack(o.lw))
    end
    o.hls = fn.h*(o.nl+o.nchs) -- total height of all lines (including choices)
    -- dxm and dym are multipliers to align (i.e. center) text
    z = o.alx
    o.dxm = (z=='l' and 0) or (z=='m' and 0.5) or (z=='r' and 1)
    z = o.alxb
    o.dxmb = (z=='l' and 0) or (z=='m' and 0.5) or (z=='r' and 1)
    z = o.alyb
    o.dymb = (z=='t' and 0) or (z=='m' and 0.5) or (z=='b' and 1)
    -- if there's a name, shift text down
    o.dyname = (o.name and fn.h) or 0
    o:setpl(true)

    return o
end

-- set various positions and lengths
function _msg:setpl(setbox)
    if setbox then
        self.wbox = self.wll+self.facedx+20 -- 20 is for margin
        self.hbox = self.hls+fn.h+self.dyname+6
    end
    self.xbox = -math.floor(self.dxmb*self.wbox)
    self.ybox = -10-math.floor(self.dymb*self.hbox)
    local x,y = self.xbox,self.ybox -- everything is wrt this
    self.xname = x+10+self.facedx
    self.yname = y+10
    if self.face then
        self.xface = x+10
        self.yface = y+math.floor(0.5*(self.hbox-self.hface))
    end
    self.xtext = x+10+self.facedx+self.dxm*(self.wbox-20-self.facedx)
    self.ytext = y+10+self.dyname
    self.xchc = x+10+self.facedx
    self.ychc = y+10+self.dyname+fn.h+self.msgdy+2
    self.xpress = x+math.floor(0.5*(self.wbox-img.press:getWidth()))
    self.ypress = y+self.hbox-math.floor(0.5*img.press:getHeight())
end

-- init message
function _msg:init()
    self.ci = 1
    self.done = false
    self.over = false
    self.once = false
    self.t0 = lt.getTime()
end

-- play message at (x,y)
function _msg:play(x,y,opt)
    -- first figure out how many chars to display
    local nc = math.round(cfg.msgspd*(lt.getTime()-self.t0),0)
    -- used for message auto-init
    if not self.once then self.once = true end
    
    -- display message box
    if opt.box then
        draw.window(x+self.xbox, y+self.ybox, self.wbox, self.hbox)
    end
    -- display name
    if self.name then
        draw.text(self.name, x+self.xname, y+self.yname, col.name)
    end
    -- display face
    if self.face then
        lg.setColor(col.white)
        lg.draw(self.face, x+self.xface, y+self.yface)
        lg.rectangle('line', x+self.xface, y+self.yface, self.wface, self.hface)
    end
    
    local i -- current bloc #
    local cl = 1 -- current line #
    local cnc = 0 -- current # chars
    local clw = self.lw[1] -- current line width
    -- display blocs
    for i = 1,#self.b do
        -- reached a new line, set new 'dx' for horizontal text alignment
        if i > self.bn[cl] then
            cl = cl+1
            clw = self.lw[cl]
        end
        if nc <= (self.b[i].nc + cnc) and not self.done then -- display the current bloc
            self.b[i]:draw(x+self.xtext-math.floor(self.dxm*clw), y+self.ytext, nc-cnc)
            self.ci = i
            break
        else
            self.b[i]:draw(x+self.xtext-math.floor(self.dxm*clw), y+self.ytext) -- display previous (completed) blocs
            cnc = cnc+self.b[i].nc
            if not self.b[i].inp and i == #self.b and not self.done then
                self.done = true
                self.t0 = lt.getTime()
                break
            end
            if self.b[i].inp and not self.b[i].done then -- keypress
                self.ci = i
                if self.b[i].tset then
                    if ((lt.getTime()-self.t0)%cfg.tblink < 0.5*cfg.tblink) and self.blinker then
                        lg.setColor(col.white)
                        lg.draw(img.press,x+self.xpress,y+self.ypress)
                    end
                    if kp == btn.ent then -- if player pressed input key
                        self.b[i].done = true
                        if i == #self.b and not self.done then
                            self.done = true
                            self.t0 = lt.getTime()
                            break
                        end
                        self.t0 = self.t0 + (lt.getTime()-self.b[i].t0)
                    end
                    return
                else
                    --[[ store time when message paused for keypress.
                    this is later used to adjust the initial time of
                    the message. --]]
                    self.b[i].t0 = lt.getTime()
                    self.b[i].tset = true
                    break
                end
            end
        end
    end

    -- press enter key to skip the message
    if not self.done and kp == btn.ent and self.skip and (not self.b[self.ci].inp or (self.b[self.ci].inp and (not self.b[self.ci].tset or self.b[self.ci].done))) then
        for i = 1,#self.b do
            if self.b[i].inp then
                self.b[i].done = true
            end
        end
        self.done = true
        self.t0 = lt.getTime()
        return
    end

    -- if message is over, don't display
    local wait = opt.wait or nil
    if self.done and not self.over then
        if self.chs then
            if not self.cchs.done then
                self.cchs:run(x+self.xchc,y+self.ychc)
            else
                self.over = true
                self.pick = self.cchs.pick
            end
        else
            if wait == nil then
                if ((lt.getTime()-self.t0)%cfg.tblink < 0.5*cfg.tblink) and self.blinker then
                    lg.setColor(col.white)
                    lg.draw(img.press,x+self.xpress,y+self.ypress)
                end
                if kp == btn.ent then self.over = true end
            elseif (lt.getTime()-self.t0 > wait) then
                self.over = true
            end
        end
    end
    
    if self.over then return end
    
end

-- plays a list of messages sequentially
function _msg.playlist(m,x,y)
    for i = 1,#m do
        if not m[i].over then
            if not m[i].once then m[i]:init() end
            m[i]:play(m[i].x or x,m[i].y or y,{wait=m[i].wait,box=m[i].box})
            break
        end
    end
end

--[[ if width 'w' is specified, this adds newlines '|n' at the
right places so that no line is greater than the specified width.
]]
function _msg.addnls(s,w)
    s = s .. ' '
    local ss = ''
    local si,sj,t
    local bi,bj,i,j = 1,1,1,1
    local t = {['|c']=10, ['|,']=2, ['|:']=2, ['|!']=2}
    b = string.unformat(s)
    --[[
    i is current position in string
    j is current position minus formatting options
    ]]
    for i = 1,s:len() do
        if s:sub(i,i) == ' ' or s:sub(i,i+1) == '|n' then
            if si and sj and fn.w(b:sub(bj,j-1)) > w then -- there must be at least one ' ' before this
                ss = ss .. s:sub(bi,si-1) .. '|n'
                bi = si+1
                bj = sj+1
            end
            si = i -- pos of the last blank space in s
            sj = j -- pos of the last blank space in s minus formatting options
        end
        -- handle when there's already a newline
        if s:sub(i,i+1) == '|n' then
            ss = ss .. s:sub(bi,i+1)
            j = j-2
            si = i+1
            sj = j
            bi = si+1
            bj = sj+1
        end
        z = s:sub(i,i+1)
        -- subtract # chars in formatting options
        if not (t[z] == nil) then
            j = j-t[z]
        end
        j = j+1
    end
    ss = ss .. s:sub(bi,s:len())
    s = ss:sub(1,ss:len()-1) -- we added a ' ' at the beginning so remove it
    return s
end

-- remove the '|!' if it's the last thing in the message
function _msg.remkp(s)
    local sl = s:len()
    if s:sub(sl-1,sl) == '|!' then s = s:sub(1,sl-2) end
    return s
end

-- replace shortcuts for color changes
function _msg.replcol(s)
    --s:gsub('|c{(%a+)}', string.format('|c%2x%2x%2x%2x', (#col['%1']==4 and unpack(col['%1'])) or unpack({col['%1'][1],col['%1'][2],col['%1'][3],255})))
    local a,b,c,d,e
    while true do
        a,b,c,d = s:find('(|c{(%a+)})')
        if a == nil then break
        else
            d = col[d]
            d = (#d==4 and d) or {d[1],d[2],d[3],255}
            s = s:gsub(c, string.format('|c%2x%2x%2x%2x',unpack(d)))
        end
    end
    return s
end

-- creates list of messages
function _msg.getms(s,wo,n,opt)
    local opt = opt or {}
    local choice = opt.choice or nil
    local nchc
    if choice then nchc = #choice else nchc = 0 end
    opt.choice = nil
    local ms,i,b,nm,a,mat = {},1,1,1
    local w = wo
    local offx
    if opt.face then
        offx = opt.face:getWidth()+10
        w = w-offx
        opt.offx = offx
    end
    if w then s = _msg.addnls(s,w) end
    local sl = s:len()
    nn = (opt.name and n-1) or n
    
    while true do
        a = string.findnth(s:sub(b,sl),'|n',nn)
        if a == nil then
            for mat in s:sub(b,sl):gmatch('|n') do
                nm = nm+1
            end
            if nm+nchc <= nn then
                opt.choice = choice
                ms[i] = _msg:new(s:sub(b,sl),nil,opt)
            else
                ms[i] = _msg:new(s:sub(b,sl),nil,opt)
                i = i+1
                opt.choice = choice
                ms[i] = _msg:new('',nil,opt)
            end
            break
        else
            a = a+b-1
            ms[i] = _msg:new(_msg.remkp(s:sub(b,a-1)),nil,opt)
            b = a+2
            i = i+1
        end
    end
    
    -- find max box width
    local mxw = ms[1].wbox
    for i = 1,#ms do
        if ms[i].wbox > mxw then mxw = ms[i].wbox end
    end
    for i = 1,#ms do
        ms[i].wbox = mxw
        ms[i].hbox = fn.h*(n+1)+6
        ms[i]:setpl()
    end
    return ms
end